#include <mpi.h>
#include <stdlib.h>
#include "domain.h"
#include "domain_output.h"

int domain_output_ppm_mpiio(domain_t *d, int var, int nxproc, int nyproc, int gridi, int gridj, char *filename) {

	static MPI_Datatype blocktype;
	static int typescreated = 0;
	MPI_File file;
	MPI_Status status;
	unsigned char *data;
	double max=-1e19, min=1e19;
	double gmax, gmin;
	double **p;
	int rank;
	int ierr;
	int nx, ny, ng;
	int i,j;
	unsigned char red, green, blue;
	int count;
	int offset;
	int rowsize, colsize;
	const int headersize=17;
	FILE *f;
	char estring[MPI_MAX_ERROR_STRING];
	int len;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	nx = d->Nx; ny = d->Ny; ng = d->Nguard;	

	rowsize = nxproc*nx; colsize = nyproc*ny;

	ierr = MPI_Type_vector(ny, 3*nx, 3*nxproc*nx, MPI_UNSIGNED_CHAR, &blocktype);
	if (ierr) fprintf(stderr,"type vector output1: %d\n",ierr);
	ierr = MPI_Type_commit(&blocktype);
	if (ierr) fprintf(stderr,"type commit output1: %d\n",ierr);
	typescreated = 1;	

	if ( (var < 0) || (var >= NVARS) ) { 
		var = 0; 
	} 

	switch (var) {
		case DENSVAR: p = d->dens; break;
		case PRESVAR: p = d->pres; break;
		case ENERVAR: p = d->ener; break;
		case TEMPVAR: p = d->temp; break;
		case MOMXVAR: p = d->momx; break;
		case MOMYVAR: p = d->momy; break;
		default: p = d->dens;
	}


	for (i=ng; i<nx+ng; i++) {
		for (j=ng; j<ny+ng; j++) {
			if ((p[i][j]) < min) min = (p[i][j]);
			if ((p[i][j]) > max) max = (p[i][j]);
		}
	}
	ierr = MPI_Allreduce( &max, &gmax, 1, MPI_DOUBLE_PRECISION, MPI_MAX, MPI_COMM_WORLD);
	ierr = MPI_Allreduce( &min, &gmin, 1, MPI_DOUBLE_PRECISION, MPI_MIN, MPI_COMM_WORLD);

	data = (unsigned char *)malloc(3 * nx * ny * sizeof(unsigned char));

	count = 0;
	for (j=ny+ng-1; j>=ng; j--) {
		for (i=ng; i<nx+ng; i++) {
			valtorgb(p[i][j],gmin,gmax,&red,&green,&blue);
			data[count] = red;   count++;
			data[count] = green; count++;
			data[count] = blue;  count++;
		    }
	}

	if (rank == 0) {
		f = fopen(filename,"w");
		fprintf(f,"P6\n%4d %4d\n255\n",rowsize,colsize);
		fclose(f);
	}	

	ierr = MPI_File_open(MPI_COMM_WORLD,filename, MPI_MODE_WRONLY | MPI_MODE_APPEND , MPI_INFO_NULL, &file);
	if (ierr) { fprintf(stderr,"RANK %d: file open output1: %d\n",rank, ierr);
			MPI_Error_string(ierr,estring,&len);
			fprintf(stderr,"%d means %s\n",ierr,estring); 
			fflush(stderr);}

	offset = headersize + (nyproc-1-gridj)*(3*rowsize*ny) + gridi*(3*nx);
	ierr = MPI_File_set_view(file, offset, blocktype, blocktype, "native", MPI_INFO_NULL);
	if (ierr) fprintf(stderr,"file setview output1: %d\n",ierr);
	ierr = MPI_File_write(file, data, 3*nx*ny, MPI_UNSIGNED_CHAR, &status);
	if (ierr) fprintf(stderr,"file write output1: %d\n",ierr);
	ierr = MPI_File_close(&file);
	if (ierr) fprintf(stderr,"file write output1: %d\n",ierr);

	return 0;
}

#ifdef PGPLOT
#include <cpgplot.h>
#endif

#define NCONTOURS 5

int domain_plot_mpi(domain_t *d) {

#ifdef PGPLOT
	int nx, ny, ng;
	double xl, xr, yl, yr, dx, dy;
	float  denscontours[NCONTOURS], prescontours[NCONTOURS];
	float *dens, *pres;
	double lmaxd=-1e19, lmind=+1e-19;
	double lmaxp=-1e19, lminp=+1e-19;
	double maxd, mind, maxp, minp;
	static int called = 0;
	int i, j, count;	
	float tr[6] = {0.,0.,0.,0.,0.,0.};
	
	ng = d->Nguard;
	nx = d->Nx;
	ny = d->Ny;

	xl = d->x[ng]; xr = d->x[nx+ng-1];
	yl = d->y[ng]; yr = d->y[ny+ng-1];
	dx = d->x[1]-d->x[0]; dy = d->y[1]-d->y[0];

	dens = (float *)malloc(nx*ny*sizeof(float));
	pres = (float *)malloc(nx*ny*sizeof(float));
	count = 0;
	for (j=0; j<ny; j++) {
		for (i=0; i<nx; i++) {
			dens[count] = (float)d->dens[i+ng][j+ng];
			pres[count] = (float)d->pres[i+ng][j+ng];
			if (dens[count] > lmaxd) lmaxd = dens[count];
			if (dens[count] < lmind) lmind = dens[count];
			if (pres[count] > lmaxp) lmaxp = pres[count];
			if (pres[count] < lminp) lminp = pres[count];
			count++;
		}
	}
	
	MPI_Allreduce(&lmaxd, &maxd, 1, MPI_DOUBLE_PRECISION, MPI_MAX, MPI_COMM_WORLD);
	MPI_Allreduce(&lmaxp, &maxp, 1, MPI_DOUBLE_PRECISION, MPI_MAX, MPI_COMM_WORLD);
	MPI_Allreduce(&lmind, &mind, 1, MPI_DOUBLE_PRECISION, MPI_MIN, MPI_COMM_WORLD);
	MPI_Allreduce(&lminp, &minp, 1, MPI_DOUBLE_PRECISION, MPI_MIN, MPI_COMM_WORLD);
	
	tr[0] = xl; tr[3] = yl;
	tr[1] = dx; tr[5] = dy;

	for (i=0; i<NCONTOURS; i++) {
		denscontours[i] = mind + (i+1)*(maxd-mind)/(NCONTOURS+1);
		prescontours[i] = minp + (i+1)*(maxp-minp)/(NCONTOURS+1);
	}


	if (!called) {
		cpgbeg(0, "/XWINDOW", 1, 1);
		called = 1;
		cpgask(0);
	}
	cpgenv(xl, xr, yl, yr, 1, 1);

	cpgsci(2);
	cpgcont(dens, nx, ny, 1, nx, 1, ny, denscontours, NCONTOURS, tr);
	if (minp != maxp) {
		cpgsci(3);
		cpgcont(pres, nx, ny, 1, nx, 1, ny, prescontours, NCONTOURS, tr);
	}
	cpgsci(1);
	
	free(dens); free(pres);
#endif
	return 0;
}

