#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include "domain.h"
#include "../util/pca_utils.h"
#include "eos.h"
#include "kurganovlevy.h"
#include "domain_output.h"
#include "boundary_conds.h"
#include "ics.h"

#define SIM_SEDOV 0
#define SIM_PROJECTILE 1

static int output = 1;

static struct option long_options[] = {
        {"output"   , no_argument, &output, 1},
        {"nooutput" , no_argument, &output, 0},
        {"npts"     , required_argument, 0, 'n'},
        {"nsteps"   , required_argument, 0, 't'},
        {"dt"       , required_argument, 0, 'd'},
        {"psize"    , required_argument, 0, 'r'},
        {"pvel"     , required_argument, 0, 'v'},
        {"pener"    , required_argument, 0, 'e'},
        {"xboundary", required_argument, 0, 'x'},
        {"yboundary", required_argument, 0, 'y'},
        {"outevery" , required_argument, 0, 'o'},
        {"simulation", required_argument, 0, 's'},
        {"help",      no_argument, 0, 'h'},
        {0, 0, 0, 0}};

int get_options(int argc, char **argv, int *npts, int *nsteps, int *xbc, int *ybc, int *simulation, int *outevery,
                                        double *dt, double *pener, double *pvel, double *psize);

int main(int argc, char **argv) {
	int npts  = 100;
	int nsteps = 500;
	double dt  = 5e-4;

	char filename[100];
	int xbc = BC_PERIODIC;
	int ybc = BC_PERIODIC;
	int outevery = 5;
	double psize = 0.1;
	double pvel = 4.;
	double pdens=50.;
	double pener=0.851072;  /* a standard test case */
	double time;
	int step;
	int simulation = SIM_PROJECTILE;
	int outputvar;
	int bcs[4];
	int outnum = 0;
	int res;

	domain_t d;
	pca_time tt;

	res = get_options(argc,argv,&npts,&nsteps,&xbc,&ybc,&simulation,&outevery, &dt, &pener, &pvel, &psize);
	if (res != 0) return 0;

	printf("Starting run with [%dx%d] grid, \n", npts, npts);
	printf("             for %d steps, each %g long \n", nsteps, dt);
	printf("             outputting every %d steps\n", outevery);
	switch(simulation) {
		case SIM_PROJECTILE:
		printf("             with projectile of size %f, velocity %f, and density %f,\n", psize, pvel, pdens);
		printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
		break;
		case SIM_SEDOV:
		printf("             with explosion energy %g in size %g,\n", pener, psize);
		printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
		break;
	}


	/* setup ICs */
	switch(simulation) {
		case SIM_SEDOV: 
			init_domain(&d, npts, npts, KL_NGUARD, -1.5, -1.5, 1.5, 1.5);
			sedov_initvalues(&d, psize, pener);
			outputvar = PRESVAR;
			break;

		case SIM_PROJECTILE: 
			init_domain(&d, npts, npts, KL_NGUARD, 0., 0., 1., 1.);
			projectile_initvalues(&d, psize, pdens, pvel);
			outputvar = DENSVAR;
			break;
	}

	/* apply boundary conditions and make thermodynamically consistant */
	bcs[0] = xbc; bcs[1] = xbc;
	bcs[2] = ybc; bcs[3] = ybc;
	apply_all_bcs(&d,bcs);
	domain_backward_dp_eos(&d);
	domain_ener_internal_to_tot(&d);  


	/* main loop */

	tick(&tt);
	if (output) domain_plot(&d);
	printf("Step\tdt\ttime\n");
	for (time=0.,step=0; step < nsteps; step++, time+=2.*dt) {
		printf("%d\t%g\t%g\n", step, dt, time);
		if (output && ((step % outevery) == 0) ) {
			sprintf(filename,"dens_test_%d.ppm", outnum); 
			domain_output_ppm(&d, outputvar, filename);
			domain_plot(&d);
			outnum++;
		}
		kl_timestep(&d, bcs, dt);
	}
	tock(&tt);

	/* final output */
	if (output) {
		sprintf(filename,"dens_test_%d.ppm", outnum); 
		domain_output_ppm(&d, outputvar,filename);
		domain_closeplot();
	}
	if (simulation == 0) domain_output_dens_r(&d,"sedov-test.dat");

	free_domain(&d);	
	return 0;
}


int get_options(int argc, char **argv, int *npts, int *nsteps, int *xbc, int *ybc, int *simulation, int *outevery,
                                        double *dt, double *pener, double *pvel, double *psize) {

	char c;
	int option_index;
	int tempint;
	double tempflt;

        while (1) {
                c = getopt_long(argc, argv, "n:t:d:r:v:x:y:o:h:e:s", long_options, &option_index);
                if (c == -1) break;

                switch (c) {
                        case 0: if (long_options[option_index].flag != 0)
                                break;

                        case 'n': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > 1000) {
                                        fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *npts);
                                  } else {
                                        *npts = tempint;
                                  }
                                  break;


                        case 't': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > 1000) {
                                        fprintf(stderr,"%s: Cannot use number of steps  %s;\n  Using %d\n", argv[0], optarg, *nsteps);
                                  } else {
                                        *nsteps  = tempint;
                                  }
                                  break;


                        case 'd': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 1.) {
                                        fprintf(stderr,"%s: Cannot use timestep %s;\n  Using %g\n", argv[0], optarg, *dt);
                                  } else {
                                        *dt = tempflt;
                                  }
                                  break;


                        case 'e': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 1.e10) {
                                        fprintf(stderr,"%s: Cannot use explosion energy %s;\n  Using %g\n", argv[0], optarg, *pener);
                                  } else {
                                        *pener = tempflt;
                                  }
                                  break;

                        case 'v': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 10.) {
                                        fprintf(stderr,"%s: Cannot use velocity %s;\n  Using %g\n", argv[0], optarg, *pvel);
                                  } else {
                                        *pvel = tempflt;
                                  }
                                  break;

                        case 'r': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 10.) {
                                        fprintf(stderr,"%s: Cannot use size %s;\n  Using %g\n", argv[0], optarg, *psize);
                                  } else {
                                        *psize = tempflt;
                                  }
                                  break;

                        case 'x': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > NBCS) {
                                        fprintf(stderr,"%s: invalid x boundary condition %s;\n  Using %d\n", argv[0], optarg, abs(*xbc));
                                  } else {
                                        *xbc = -tempint;
                                  }
                                  break;

                        case 'y': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > NBCS) {
                                        fprintf(stderr,"%s: invalid y boundary condition %s;\n  Using %d\n", argv[0], optarg, abs(*ybc));
                                  } else {
                                        *ybc = -tempint;
                                  }
                                  break;

                        case 'o': tempint = atoi(optarg);
                                  if (tempint < 1) {
                                        fprintf(stderr,"%s: invalid output frequency %s;\n  Using %d\n", argv[0], optarg, *outevery);
                                  } else {
                                        *outevery = tempint;
                                  }
                                  break;

                        case 's': tempint = atoi(optarg);
                                  if (tempint != SIM_PROJECTILE && tempint != SIM_SEDOV ) {
                                        fprintf(stderr,"%s: invalid simulation number %s;\n  Using %d\n", argv[0], optarg, *simulation);
                                  } else {
                                        *simulation = tempint;
                                  }
                                  break;

			case 'h': 
				  puts("Options: ");
				  puts("    --npts=N      (-n N): Set the number of grid cells in each direction.");
				  puts("    --nsteps=N    (-t N): Set the number of time steps taken.");
				  puts("    --dt=X        (-d X): Set the time step.");
				  puts("    --pvel=X      (-v X): Set the velocity parameter (projectile run).");
				  puts("    --pener=X     (-v X): Set the explosion parameter (Sedov run).");
				  puts("    --psize=X     (-r X): Set the size parameter.");
				  puts("    --simulation=N(-s N): Set the simulation; 0 == Sedov, 1 == projectile.");
				  puts("    --outevery=N  (-o N): Set output every N timesteps.");
				  puts("    --xboundary=N (-x N): Set the x-boundary condition: ");
				  puts("    --yboundary=N (-y N): Set the y-boundary condition: ");
				  puts("                          1=outflow, 2=reflecting, 3=periodic");
				  puts("");
 				  return +1;

			default: printf("Invalid option %s\n", optarg);
				 break;
		}
	}
	return 0;
}
