#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "kurganovlevy.h"
#include "row.h"
#include "eos.h"
#include "reconstruction.h"
#include "boundary_conds.h"

/*
 *  Implements a 1.5-d dimensionally split Kurganov-Levy hydro solver (eg, 
 *  carries around a passive tangential velocity).  To be used in a 
 *  Strang-split 2d solver.
 *
 *  Ref: http://arxiv.org/abs/math.NA/0002133
 */


/* define some floors */
#define KL_MINE 1.e-13
#define KL_MIND 1.e-13
#define KL_MAXE 1.e+29
#define KL_MAXD 1.e+29

double max(double x, double y) {
	return (x > y ? x : y);	
}


int calcflux(row_t *u, row_t *f) {
	
	int i, mustflip = 0;
	double vx;

	if (f->N < u->N) {
		printf("calcflux: sizes don't match.  Failing\n");	
		return -1;
	}
	
	if (u->enerflag != FLAG_TOTALE) mustflip = 1;
	if (mustflip) ener_internal_to_tot(u);
	for (i=0; i<u->N; i++) {
		vx = u->momx[i]/u->dens[i];

		f->dens[i] = u->momx[i];
		f->momx[i] = u->momx[i]*vx + u->pres[i];
		f->momy[i] = u->momy[i]*vx;
		f->ener[i] = (u->ener[i] + u->pres[i])*vx;
	}
	if (mustflip) ener_tot_to_internal(u);

	return 0;
}

int kl_semidiscrete_row(row_t *r, int nguard, double *ddensdt, double *dmomxdt, double *dmomydt, double *denerdt) {
                                      

	/* assume uniform grid */
        double dx = r->x[1]-r->x[0];
	double gamma;
	
	int i,n;

	row_t leftint;
	row_t rightint;
	row_t fluxleft, fluxright;
	double *csleft, *csright, *csint;

	int status;

	n = r->N;
	gamma = eos_gamma();

	/* there is one more interface than cell in a 1d array */
	status = init_row(&leftint,  n+1, r->x[0]-dx/2., r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&rightint, n+1, r->x[0]-dx/2., r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&fluxleft, n+1, r->x[0]-dx/2,  r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&fluxright,n+1, r->x[0]-dx/2,  r->x[n-1]+dx/2., FLAG_TOTALE);

	if (status != 0) {
		fprintf(stderr,"Could not create rows for kl_semidiscrete; failing.\n");
		return -1;
	}
	
	csleft  = (double *)malloc((n+1) * sizeof(double));
	csright = (double *)malloc((n+1) * sizeof(double));
	
	/* csint[i] is the sound speed at the interface between cell i-1 and i */
	csint   = (double *)malloc((n+1) * sizeof(double));

	if (csleft == NULL || csright == NULL || csint == NULL) {
		fprintf(stderr,"Could not malloc sound speed arrays; failing.\n");
		free_row(&leftint); free_row(&rightint); free_row(&fluxleft); free_row(&fluxright);
		return -1;
	}

	/* find interface values of primitive variables */
	/* interface i is between cell i-1 and i; interpolated from right and left */
	status = intvalues(n, r->dens, r->x, rightint.dens, leftint.dens);
	status+= intvalues(n, r->momx, r->x, rightint.momx, leftint.momx);
	status+= intvalues(n, r->momy, r->x, rightint.momy, leftint.momy);
	status+= intvalues(n, r->ener, r->x, rightint.ener, leftint.ener);
	if (status != 0) {
		fprintf(stderr,"kl_semidiscrete cound not interpolate values; failing.\n");
		return -2;
	}

	limit(n, rightint.dens, KL_MIND, KL_MAXD);
	limit(n, leftint .dens, KL_MIND, KL_MAXD);

	ener_tot_to_internal(&leftint);  limit(n, leftint.ener, KL_MINE, KL_MAXE);
	ener_tot_to_internal(&rightint); limit(n, rightint.ener, KL_MINE, KL_MAXE);

        forward_eos_row(&leftint);
        forward_eos_row(&rightint);

	/* find maximum signal speeds */
	for (i=0; i<n+1; i++) {
		csleft[i] = sqrt(gamma * leftint.pres[i] /leftint.dens[i]);
		csright[i]= sqrt(gamma * rightint.pres[i]/rightint.dens[i]);

		csleft[i] = csleft[i] + fabs(leftint .momx[i]/leftint.dens[i]);
		csright[i]= csright[i]+ fabs(rightint.momx[i]/rightint.dens[i]);
	}

	for (i=0; i<n+1; i++) {
		csint[i] = max(csright[i],csleft[i]);
	}

	calcflux(&leftint, &fluxleft);
	calcflux(&rightint,&fluxright);

	for (i=0; i<n-1; i++) {
		ddensdt[i] = -(fluxright.dens[i+1] + fluxleft.dens[i+1] - fluxright.dens[i] - fluxleft.dens[i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.dens[i+1] - leftint.dens[i+1]) - csint[i]/(2.*dx) * (rightint.dens[i] - leftint.dens[i]);
		dmomxdt[i] = -(fluxright.momx[i+1] + fluxleft.momx[i+1] - fluxright.momx[i] - fluxleft.momx[i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.momx[i+1] - leftint.momx[i+1]) - csint[i]/(2.*dx) * (rightint.momx[i] - leftint.momx[i]);
		dmomydt[i] = -(fluxright.momy[i+1] + fluxleft.momy[i+1] - fluxright.momy[i] - fluxleft.momy[i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.momy[i+1] - leftint.momy[i+1]) - csint[i]/(2.*dx) * (rightint.momy[i] - leftint.momy[i]);
		denerdt[i] = -(fluxright.ener[i+1] + fluxleft.ener[i+1] - fluxright.ener[i] - fluxleft.ener[i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.ener[i+1] - leftint.ener[i+1]) - csint[i]/(2.*dx) * (rightint.ener[i] - leftint.ener[i]);
	}

	/* don't update the guardcells */
	for (i=0; i<nguard; i++) {
		ddensdt[i] = 0.;
		ddensdt[n-nguard+i] = 0.;
		dmomxdt[i] = 0.;
		dmomxdt[n-nguard+i] = 0.;
		dmomydt[i] = 0.;
		dmomydt[n-nguard+i] = 0.;
		denerdt[i] = 0.;
		denerdt[n-nguard+i] = 0.;
	}


	free_row(&leftint);
	free_row(&rightint);
	free_row(&fluxleft);
	free_row(&fluxright);
	free(csleft); free(csright); free(csint);

	return 0;
}


/*
 * Does a 2nd order RK time integration of the hydro equations, using 
 * the semidiscrete routine to get an instantaneous dX/dt.
 *
 * Could do 3rd order RK but the Strang splitting makes us at best
 * 2nd order in time 
 */
int kl_row_rk2(row_t *r, double dt) {

	row_t rl;
	int n, status;
	int i;
	double *ddensdt, *dmomxdt, *dmomydt, *denerdt;

	n = r->N;

	ddensdt = (double *)malloc(n * sizeof(double));
	dmomxdt = (double *)malloc(n * sizeof(double));
	dmomydt = (double *)malloc(n * sizeof(double));
	denerdt = (double *)malloc(n * sizeof(double));

	if (ddensdt == NULL || dmomxdt == NULL || dmomydt == NULL || denerdt == 0) {
		fprintf(stderr,"kurganovlevy_row_rk2: cannot allocate arrays for derivs.  Failing\n");
		return -2;
	}

	/* eos of input data and ensure in valid range */
	ener_tot_to_internal(r);
	limit(n, r->ener, KL_MINE, KL_MAXE);
        forward_eos_row(r);
	ener_internal_to_tot(r);
	
	/* intermediate stage data needed for RK step */	
	status = init_row(&rl, n, r->x[0], r->x[n-1], FLAG_TOTALE);
	if (status != 0) {
		fprintf(stderr,"kurganovlevy_row_rk2; cannot initialize intermediate row.  Failing\n");
		free(ddensdt); free(dmomxdt); free(dmomydt); free(denerdt);
		return -2;
	}

	/* first stage of RK; get dX/dt at current timestep */
	kl_semidiscrete_row(r, KL_NGUARD, ddensdt, dmomxdt, dmomydt, denerdt);

	for (i=0; i<n; i++) {
		rl.dens[i] = r->dens[i] + dt*ddensdt[i];
		rl.momx[i] = r->momx[i] + dt*dmomxdt[i];
		rl.momy[i] = r->momy[i] + dt*dmomydt[i];
		rl.ener[i] = r->ener[i] + dt*denerdt[i];
	}

	limit(n, rl.dens, KL_MIND, KL_MAXD);
	ener_tot_to_internal(&rl);
	limit(n, rl.ener, KL_MINE, KL_MAXD);
	forward_eos_row(&rl);
	ener_internal_to_tot(&rl);

	/* predictor-corrector type step */
	kl_semidiscrete_row(&rl, KL_NGUARD, ddensdt, dmomxdt, dmomydt, denerdt);
	for (i=0; i<n; i++) {
		rl.dens[i] = 0.5*r->dens[i] + (1.-0.5)*(rl.dens[i] + dt*ddensdt[i]);
		rl.momx[i] = 0.5*r->momx[i] + (1.-0.5)*(rl.momx[i] + dt*dmomxdt[i]);
		rl.momy[i] = 0.5*r->momy[i] + (1.-0.5)*(rl.momy[i] + dt*dmomydt[i]);
		rl.ener[i] = 0.5*r->ener[i] + (1.-0.5)*(rl.ener[i] + dt*denerdt[i]);
	}

	limit(n, rl.dens, KL_MIND, KL_MAXD);
	ener_tot_to_internal(&rl);
	limit(n, rl.ener, KL_MINE, KL_MAXD);
	forward_eos_row(&rl);
	ener_internal_to_tot(&rl);

	/* copy new data back into old row */
	memcpy(r->dens, rl.dens, n*sizeof(double));
	memcpy(r->momx, rl.momx, n*sizeof(double));
	memcpy(r->momy, rl.momy, n*sizeof(double));
	memcpy(r->ener, rl.ener, n*sizeof(double));
	memcpy(r->pres, rl.pres, n*sizeof(double));
	memcpy(r->temp, rl.temp, n*sizeof(double));

	free_row(&rl);
	free(ddensdt);
	free(dmomxdt);
	free(dmomydt);
	free(denerdt);

	return 0;
}
