#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * These routines implement 3rd order 1d-CWENO reconstructions assuming a uniform grid.
 */ 

int limit(int n, double *v, double min, double max) {
	int i;
	
	for (i=0; i<n; i++) {
		if (v[i] < min) v[i] = min;
	}
	for (i=0; i<n; i++) {
		if (v[i] > max) v[i] = max;
	}

	return 0;
}

int reconstruction(int n, double *u, double *x, double *wl, double *wc, double *wr, double p, double *a, double *b, double *c) {

	double eps = 1.e-9;
	double dx = x[1]-x[0];
	double isl, isc, isr;
	double alphal, alphac, alphar, alphatot;

	int i;

	for (i=1; i<n-1; i++) {
		isl = pow((u[i] - u[i-1]),2);
		isc = 13./3.*pow(( u[i+1] - 2.*u[i] + u[i-1] ),2) + 0.25*pow((u[i+1] - u[i-1]),2);
		isr = pow((u[i+1] - u[i]),2);

		alphal = 0.25  / pow(eps + isl,p);
		alphac = 0.50  / pow(eps + isc,p);
		alphar = 0.25  / pow(eps + isr,p);

		alphatot = alphar+alphac+alphal;

		wl[i] = alphal/alphatot;
		wr[i] = alphar/alphatot;
		wc[i] = alphac/alphatot;
	}

	a[0]   = u[0];   b[0]   = 0; c[0]   = 0;
	a[n-1] = u[n-1]; b[n-1] = 0; c[n-1] = 0;

	for (i=1; i<n-1; i++) {
		a[i] = u[i] - wc[i]/12.*(u[i+1] - 2*u[i] + u[i-1]);
		b[i] = (wr[i]*(u[i+1] - u[i]) + wc[i]/2.*(u[i+1] - u[i-1]) + wl[i]*(u[i] - u[i-1]))/dx;
		c[i] = 2.*wc[i]/(dx*dx) * ( u[i-1] - 2.*u[i] + u[i+1]);
		if (isnan(a[i])) {fprintf(stderr,"a[%d] = NaN; u[i] = %g, wc[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wc[i], u[i+1], u[i-1]);}
		if (isnan(b[i])) {fprintf(stderr,"b[%d] = NaN; u[i] = %g, wr[i] = %g, wc[i] = %g, wl[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wr[i], wc[i], wl[i], u[i+1], u[i-1]);}
		if (isnan(c[i])) {fprintf(stderr,"c[%d] = NaN; u[i] = %g, wr[i] = %g, wc[i] = %g, wl[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wr[i], wc[i], wl[i], u[i+1], u[i-1]);}
	}

	return 0;
}

/*
 * interpolate to interfaces from the right and left.
 */
int intvalues(int n, double *u, double *x, double *uhalfright, double *uhalfleft) {

	double *wl, *wc, *wr;
	double *a, *b, *c;
	double dx = x[1] - x[0];
	int i, status;

	wr = (double *)malloc(n * sizeof(double));
	wc = (double *)malloc(n * sizeof(double));
	wl = (double *)malloc(n * sizeof(double));
	if (wr == NULL || wc == NULL || wl == NULL) {
		fprintf(stderr,"intvalues: could not allocate 1d arrays; failing\n");
		return -2;
	}

	a = (double *)malloc(n * sizeof(double));
	b = (double *)malloc(n * sizeof(double));
	c = (double *)malloc(n * sizeof(double));
	if (a == NULL || b == NULL || c == NULL) {
		fprintf(stderr,"intvalues: could not allocate 1d arrays; failing\n");
		free(wr); free(wc); free(wl);
		return -2;
	}

	status = reconstruction(n, u, x, wl, wc, wr, 2., a, b, c);

	free(wr); free(wc); free(wl);

	/* interpolate to the ith boundary (between cell i-1 and i) from the right and left */
	for (i=0; i<n; i++) {
		uhalfright[i]   = a[i] - dx/2.*b[i] + dx*dx/8.*c[i];
		uhalfleft [i+1] = a[i] + dx/2.*b[i] + dx*dx/8.*c[i];
	}
	uhalfright[n]  = u[n-1];
	uhalfleft [0]  = u[0];
	
	free(a); free(b); free(c);

	return 0;
}
