#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include <getopt.h>
#include <time.h>
#include "domain.h"
#include "util.h"
#include "eos.h"
#include "kurganovlevy_mpi.h"
#include "domain_output.h"
#include "domain_output_mpio.h"
#include "boundary_conds_mpi.h"
#include "ics.h"

#define SIM_SEDOV 0
#define SIM_PROJECTILE 1

#define MASTER_TASK 0

int get_options(int argc, char **argv, int *npts, int *nsteps, int *xbc, int *ybc, int *simulation, int *outevery,
                                        int *output, int *nxproc, int *nyproc,
                                        double *cfl, double *pener, double *pvel, double *psize);

int main(int argc, char **argv) {
	int npts  = 100;
	int nxpts, nypts;
	int nsteps = 500;
	double dt  = 5.e-4;
	double locdt;
	double cfl = 0.375;

	char filename[100];
	int xbc = BC_PERIODIC;
	int ybc = BC_PERIODIC;
	int outevery = 5;
	int ierr;
	double psize = 0.1;
	double pvel = 4.;
	double pdens=50.;
	double pener=0.851072;  /* a standard test case */
	double time;
	int step;
	int simulation = SIM_PROJECTILE;
	int outputvar;
	int bcs[4];
	int outnum = 0;
	double globxmax, globymax, globxmin, globymin;  
	int rank, size;     /* which MPI task we are, and how many there are */
	int nxproc=1, nyproc=1;
	int output=1;
	int gridi, gridj;   /* where we are in a 2d decomposition */
	double locxmax, locymax, locxmin, locymin;  
	int xlneigh, xrneigh, ylneigh, yrneigh;
	double dx, dy;

	domain_t d;

	double dparams[4];
	int    iparams[9];

	struct timeval tt;
  	double timemin, timemean, timemax;

	ierr = MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD,&size);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);

	if (rank == MASTER_TASK) { 
		ierr = get_options(argc,argv,&npts,&nsteps,&xbc,&ybc,&simulation,&outevery, &output, &nxproc, &nyproc, &cfl, &pener, &pvel, &psize);
		if (ierr) npts = -1;
		iparams[0] = npts; 
		iparams[1] = nsteps;
		iparams[2] = xbc;  
		iparams[3] = ybc;
		iparams[4] = simulation; 
		iparams[5] = outevery; 
		iparams[6] = output; 
		iparams[7] = nxproc; 
		iparams[8] = nyproc;
		dparams[0] = cfl; 
		dparams[1] = pener;
		dparams[2] = pvel;
		dparams[3] = psize;
	}
	ierr = MPI_Bcast(iparams, 9, MPI_INTEGER, MASTER_TASK, MPI_COMM_WORLD);
	ierr = MPI_Bcast(dparams, 4, MPI_DOUBLE_PRECISION, MASTER_TASK, MPI_COMM_WORLD);
	if (rank != MASTER_TASK) {
		npts = iparams[0];
		nsteps = iparams[1];
		xbc = iparams[2];
		ybc = iparams[3];
		simulation = iparams[4];
		outevery = iparams[5];
		output  = iparams[6];
		nxproc = iparams[7];
		nyproc = iparams[8];
		cfl = dparams[0];
		pener = dparams[1];
		pvel = dparams[2];
		psize = dparams[3];
	}
	if (npts == -1) {
		MPI_Finalize();
		return 0;
	}

	if (nxproc * nyproc != size) {
		fprintf(stderr,"Error: number of MPI tasks %d should equal nxproc (%d) x nyproc (%d)\n", size, nxproc, nyproc);
		MPI_Finalize();
		return 0;
	}

	gridi = rank % nxproc;
	gridj = rank / nxproc;

	nxpts = npts / nxproc;
	nypts = npts / nyproc;
	

	if (rank == MASTER_TASK) {
		printf("Starting run with [%dx%d] ([%d,%d]) grid, \n", npts, npts, nxpts, nypts);
		printf("             for %d steps\n", nsteps);
		printf("             outputting every %d steps\n", outevery);
	}
	switch(simulation) {
		case SIM_PROJECTILE:
			globxmin = 0.; globxmax = +1.;
			globymin = 0.; globymax = +1.;
			if (rank == MASTER_TASK) {
				printf("             with projectile of size %f, velocity %f, and density %f,\n", psize, pvel, pdens);
				printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
			}
			break;
		case SIM_SEDOV:
			globxmin = -1.5; globxmax = +1.5;
			globymin = -1.5; globymax = +1.5;
			if (rank == MASTER_TASK) {
				printf("             with explosion energy %g in size %g,\n", pener, psize);
				printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
			}
			break;
	}

	dx = (globxmax-globxmin)/(nxproc*nxpts-1);
	dy = (globymax-globymin)/(nyproc*nypts-1);

	locxmin = globxmin + dx*nxpts*gridi;
	locxmax = globxmin + dx*(nxpts*(gridi+1)-1);
	locymin = globymin + dy*nypts*gridj;
	locymax = globymin + dy*(nypts*(gridj+1)-1);

	init_domain(&d, nxpts, nypts, KL_NGUARD, locxmin, locymin, locxmax, locymax);

	/* setup ICs */
	switch(simulation) {
		case SIM_SEDOV: 
			sedov_initvalues(&d, psize, pener);
			outputvar = PRESVAR;
			break;

		case SIM_PROJECTILE: 
			projectile_initvalues(&d, psize, pdens, pvel);
			outputvar = DENSVAR;
			break;
	}

	/* figure out where are neighbors are or if they are external boundaries */
	if (gridi == 0       ) xlneigh = xbc; else xlneigh = rank-1;
	if (gridi == nxproc-1) xrneigh = xbc; else xrneigh = rank+1;
	if (gridj == 0       ) ylneigh = ybc; else ylneigh = rank - nxproc;
	if (gridj == nyproc-1) yrneigh = ybc; else yrneigh = rank + nxproc;

	if ((xlneigh == BC_PERIODIC) && (nxproc > 1)) xlneigh = rank+nxproc-1;
	if ((xrneigh == BC_PERIODIC) && (nxproc > 1)) xrneigh = rank-nxproc+1;
	if ((ylneigh == BC_PERIODIC) && (nyproc > 1)) ylneigh = rank+(nyproc-1)*nxproc;
	if ((yrneigh == BC_PERIODIC) && (nyproc > 1)) yrneigh = rank-(nyproc-1)*nxproc;

	printf("Task %d: local domain is [%g,%g]x[%g,%g], neighbors are (%d,%d,%d,%d)\n", rank, locxmin, locxmax, locymin, locymax, xlneigh, xrneigh, ylneigh, yrneigh);

	/* apply boundary conditions and make thermodynamically consistant */
	bcs[0] = xlneigh; bcs[1] = xrneigh;
	bcs[2] = ylneigh; bcs[3] = yrneigh;
	apply_all_bcs(&d,bcs,gridi,gridj);
	domain_backward_dp_eos(&d);
	domain_ener_internal_to_tot(&d);  

	/* get initial timestep */
	domain_timestep(&d, &locdt);
	MPI_Allreduce( &locdt, &dt, 1, MPI_DOUBLE_PRECISION, MPI_MIN, MPI_COMM_WORLD);
	dt *= cfl;

	/* main loop */

        tick(&tt);
	if (rank == MASTER_TASK && output) printf("Step\tdt\ttime\n");
	for (time=0.,step=0, outnum=0; step < nsteps; step++, time+=2.*dt) {
		if (rank == MASTER_TASK && output) printf("%d\t%g\t%g\n", step, dt, time);
		if (output && ((step % outevery) == 0) ) {
			sprintf(filename,"dens_test_%d.ppm", outnum); 
			domain_output_ppm_mpiio(&d, outputvar, nxproc, nyproc, gridi, gridj, filename);
			domain_plot_mpi(&d);
			outnum++;
		}
		kl_timestep(&d, bcs, dt, gridi, gridj);
	
		domain_timestep(&d, &locdt);
		MPI_Allreduce( &locdt, &dt, 1, MPI_DOUBLE_PRECISION, MPI_MIN, MPI_COMM_WORLD);
		dt *= cfl;
	}
        mpitock(&tt, &timemin, &timemean, &timemax);

	/* final output */
	if (output) {
		sprintf(filename,"dens_test_%d.ppm", outnum); 
		domain_output_ppm_mpiio(&d, outputvar, nxproc, nyproc, gridi, gridj, filename);
		domain_closeplot();
	}
	if (simulation == 0) domain_output_dens_r(&d,"sedov-test.dat");
	if (rank == MASTER_TASK) {
		fprintf(stdout,"Loop timings by process: min/mean/max = %g %g %g\n", 
 		        timemin, timemean, timemax);
	}

	free_domain(&d);	
	MPI_Finalize();
	return 0;
}


int get_options(int argc, char **argv, int *npts, int *nsteps, int *xbc, int *ybc, int *simulation, int *outevery,
				  	int *output, int *nxproc, int *nyproc,
                                        double *cfl, double *pener, double *pvel, double *psize) {
static int myoutput = 1;
static struct option long_options[] = {
        {"output"   , no_argument, &myoutput, 1},
        {"nooutput" , no_argument, &myoutput, 0},
        {"npts"     , required_argument, 0, 'n'},
        {"nsteps"   , required_argument, 0, 't'},
        {"cfl"      , required_argument, 0, 'c'},
        {"psize"    , required_argument, 0, 'r'},
        {"pvel"     , required_argument, 0, 'v'},
        {"pener"    , required_argument, 0, 'e'},
        {"xboundary", required_argument, 0, 'x'},
        {"yboundary", required_argument, 0, 'y'},
        {"outevery" , required_argument, 0, 'o'},
        {"simulation", required_argument, 0, 's'},
        {"nxproc", required_argument, 0, 'X'},
        {"nyproc", required_argument, 0, 'Y'},
        {"help",      no_argument, 0, 'h'},
        {0, 0, 0, 0}};


	char c;
	int option_index;
	int tempint;
	double tempflt;

        while (1) {
                c = getopt_long(argc, argv, "n:t:d:r:v:x:y:o:h:e:s:X:Y", long_options, &option_index);
                if (c == -1) break;

                switch (c) {
                        case 0: if (long_options[option_index].flag != 0)
                                break;

                        case 'n': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > 7500) {
                                        fprintf(stderr,"%s: Cannot use number of points %s;\n  Using %d\n", argv[0], optarg, *npts);
                                  } else {
                                        *npts = tempint;
                                  }
                                  break;


                        case 't': tempint = atoi(optarg);
                                  if (tempint < 1 ) {
                                        fprintf(stderr,"%s: Cannot use number of steps  %s;\n  Using %d\n", argv[0], optarg, *nsteps);
                                  } else {
                                        *nsteps  = tempint;
                                  }
                                  break;


                        case 'c': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 0.5) {
                                        fprintf(stderr,"%s: Cannot use cfl parameter %s;\n  Using %g\n", argv[0], optarg, *cfl);
                                  } else {
                                        *cfl = tempflt;
                                  }
                                  break;


                        case 'e': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 1.e10) {
                                        fprintf(stderr,"%s: Cannot use explosion energy %s;\n  Using %g\n", argv[0], optarg, *pener);
                                  } else {
                                        *pener = tempflt;
                                  }
                                  break;

                        case 'v': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 10.) {
                                        fprintf(stderr,"%s: Cannot use velocity %s;\n  Using %g\n", argv[0], optarg, *pvel);
                                  } else {
                                        *pvel = tempflt;
                                  }
                                  break;

                        case 'r': tempflt = atof(optarg);
                                  if (tempflt < 0. || tempflt > 10.) {
                                        fprintf(stderr,"%s: Cannot use size %s;\n  Using %g\n", argv[0], optarg, *psize);
                                  } else {
                                        *psize = tempflt;
                                  }
                                  break;

                        case 'x': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > NBCS) {
                                        fprintf(stderr,"%s: invalid x boundary condition %s;\n  Using %d\n", argv[0], optarg, abs(*xbc));
                                  } else {
                                        *xbc = -tempint;
                                  }
                                  break;

                        case 'y': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > NBCS) {
                                        fprintf(stderr,"%s: invalid y boundary condition %s;\n  Using %d\n", argv[0], optarg, abs(*ybc));
                                  } else {
                                        *ybc = -tempint;
                                  }
                                  break;

                        case 'o': tempint = atoi(optarg);
                                  if (tempint < 1) {
                                        fprintf(stderr,"%s: invalid output frequency %s;\n  Using %d\n", argv[0], optarg, *outevery);
                                  } else {
                                        *outevery = tempint;
                                  }
                                  break;

                        case 's': tempint = atoi(optarg);
                                  if (tempint != SIM_PROJECTILE && tempint != SIM_SEDOV ) {
                                        fprintf(stderr,"%s: invalid simulation number %s;\n  Using %d\n", argv[0], optarg, *simulation);
                                  } else {
                                        *simulation = tempint;
                                  }
                                  break;

                        case 'X': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > 1000) {
                                        fprintf(stderr,"%s: invalid number of processors in X direction %s;\n  Using %d\n", argv[0], optarg, *nxproc);
                                  } else {
                                        *nxproc = tempint;
                                  }
                                  break;


                        case 'Y': tempint = atoi(optarg);
                                  if (tempint < 1 || tempint > 1000) {
                                        fprintf(stderr,"%s: invalid number of processors in Y direction %s;\n  Using %d\n", argv[0], optarg, *nyproc);
                                  } else {
                                        *nyproc = tempint;
                                  }
                                  break;

			case 'h': 
				  puts("Options: ");
				  puts("    --npts=N      (-n N): Set the number of grid cells in each direction.");
				  puts("    --nsteps=N    (-t N): Set the number of time steps taken.");
				  puts("    --cfl=X       (-c X): Set the CFL time step.");
				  puts("    --pvel=X      (-v X): Set the velocity parameter (projectile run).");
				  puts("    --pener=X     (-v X): Set the explosion parameter (Sedov run).");
				  puts("    --psize=X     (-r X): Set the size parameter.");
				  puts("    --simulation=N(-s N): Set the simulation; 0 == Sedov, 1 == projectile.");
				  puts("    --outevery=N  (-o N): Set output every N timesteps.");
				  puts("    --xboundary=N (-x N): Set the x-boundary condition: ");
				  puts("    --yboundary=N (-y N): Set the y-boundary condition: ");
				  puts("                          1=outflow, 2=reflecting, 3=periodic");
				  puts("    --nxproc=N    (-X N): Set the number of processors in the X direction.");
				  puts("    --nyproc=N    (-Y N): Set the number of processors in the Y direction.");
				  puts("");
 				  return +1;

			default: printf("Invalid option %s\n", optarg);
				 break;
		}
	}
        *output = myoutput;
	return 0;
}
