#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "domain.h"

int init_domain(domain_t *d, int nx, int ny, int nguard, double startx, double starty, double endx, double endy) {

	double *tmp   = NULL;
	long int i;
	int nxt, nyt;
	long int arrsize;

	if (nx < 0 || nx > MAX_DOMAINSIZE) {
		fprintf(stderr,"init_domain: Invalid x size of domain %d\n", nx);
		return -4;
	}
	if (ny < 0 || ny > MAX_DOMAINSIZE) {
		fprintf(stderr,"init_domain: Invalid y size of domain %d\n", ny);
		return -4;
	}

	nxt = nx + 2*nguard;
	nyt = ny + 2*nguard;
	d->Nx = nx;
	d->Ny = ny;
	d->Nguard = nguard;

	d->x = (double *)malloc(nxt * sizeof(double));
	d->y = (double *)malloc(nyt * sizeof(double));

	d->data = (double ***)malloc(NVARS * sizeof(double **));
	d->data[DENSVAR] = (double **)malloc(nxt * sizeof(double *));
	d->data[PRESVAR] = (double **)malloc(nxt * sizeof(double *));
	d->data[TEMPVAR] = (double **)malloc(nxt * sizeof(double *));
	d->data[ENERVAR] = (double **)malloc(nxt * sizeof(double *));
	d->data[MOMXVAR] = (double **)malloc(nxt * sizeof(double *));
	d->data[MOMYVAR] = (double **)malloc(nxt * sizeof(double *));

	if (d->x == NULL || d->y == NULL || d->data == NULL ) {
		fprintf(stderr,"init_domain: Could not allocate memory for 1d arrays; what is going on?\n");
		free(d->x); free(d->y); free(d->data[DENSVAR]); 
		return -1;
	}

	for (i=0; i<nxt; i++) d->x[i] = startx + (i-nguard)*(endx-startx)/((double)(nx-1));
	for (i=0; i<nyt; i++) d->y[i] = starty + (i-nguard)*(endy-starty)/((double)(ny-1));

	arrsize = nxt*nyt;
	
	tmp = (double *)malloc(NVARS * arrsize * sizeof(double));
	if (tmp == NULL) {
		fprintf(stderr,"init_domain: Could not allocate memory for unknown values.\n");
		free(d->x); free(d->y); free(d->data[DENSVAR]); 
		return -2;
	}

	for (i=0; i<nxt; i++) {
		(d->data)[DENSVAR][i] = &(tmp[arrsize*DENSVAR + i*nyt]);
		(d->data)[ENERVAR][i] = &(tmp[arrsize*ENERVAR + i*nyt]);
		(d->data)[PRESVAR][i] = &(tmp[arrsize*PRESVAR + i*nyt]);
		(d->data)[TEMPVAR][i] = &(tmp[arrsize*TEMPVAR + i*nyt]);
		(d->data)[MOMXVAR][i] = &(tmp[arrsize*MOMXVAR + i*nyt]);
		(d->data)[MOMYVAR][i] = &(tmp[arrsize*MOMYVAR + i*nyt]);
	}

	d->enerflag = FLAG_TOTALE;
	d->alloced = (short int)1;
	return 0;
}


int domain_ener_tot_to_internal(domain_t *d) {

	int i,j;
	double ke;
	int nxt, nyt;

	nxt = d->Nx + 2*d->Nguard;
	nyt = d->Ny + 2*d->Nguard;

	if (d->enerflag == FLAG_SPECIFICINTERNALE) return 0;

	for (i=0; i<nxt; i++) {
		for (j=0; j<nyt; j++) {
			ke = 0.5*( (d->data[MOMXVAR][i][j])*(d->data[MOMXVAR][i][j]) + (d->data[MOMYVAR][i][j])*(d->data[MOMYVAR][i][j]) )/(d->data[DENSVAR][i][j]);
			d->data[ENERVAR][i][j] = (d->data[ENERVAR][i][j] - ke)/(d->data[DENSVAR][i][j]);
		}
	}

	d->enerflag = FLAG_SPECIFICINTERNALE;
	return 0;
}


int domain_ener_internal_to_tot(domain_t *d) {

	int i,j;
	double ke;
	int nxt, nyt;

	nxt = d->Nx + 2*d->Nguard;
	nyt = d->Ny + 2*d->Nguard;

	if (d->enerflag == FLAG_TOTALE) return 0;

	for (i=0; i<nxt; i++) {
		for (j=0; j<nyt; j++) {
			ke = 0.5*( (d->data[MOMXVAR][i][j])*(d->data[MOMXVAR][i][j]) + (d->data[MOMYVAR][i][j])*(d->data[MOMYVAR][i][j]) )/(d->data[DENSVAR][i][j]);
			d->data[ENERVAR][i][j] = d->data[ENERVAR][i][j]*d->data[DENSVAR][i][j] + ke;
		}
	}

	d->enerflag = FLAG_TOTALE;
	return 0;
}



int free_domain(domain_t *d){

	if (d->alloced) {
		d->Nx = 0;	
		d->Ny = 0;	
		d->Nguard = 0;

		free(d->x); free(d->y);

		free(d->data[DENSVAR][0]);	

		free(d->data[DENSVAR]);	
		free(d->data[PRESVAR]);	
		free(d->data[ENERVAR]);	
		free(d->data[TEMPVAR]);	
		free(d->data[MOMXVAR]);	
		free(d->data[MOMYVAR]);	

		d->alloced = (short int)0;
	}
	return 0;
}

int domain_extract_row(domain_t *d, row_t *r, int rowcol, int direction) {

	int n;
	int status;
	int nxt, nyt;

	nxt = d->Nx + 2*d->Nguard;
	nyt = d->Ny + 2*d->Nguard;

	if (direction != XDIR && direction != YDIR) {
		fprintf(stderr,"extract_row: invalid direction %d; failing\n", direction);
		return -1;
	}

	if (direction == XDIR) {
		if (rowcol < 0 || rowcol >= nyt) {
			fprintf(stderr,"extract_row: invalid row %d; failing\n", rowcol);
			return -1;
		}

		if ((r->alloced != 0) && (r->N != nxt)) free_row(r);
		if (r->alloced == 0) {
			fprintf(stderr,"Have to re-allocate xrow.\n");
			status = init_row(r, nxt, d->x[0], d->x[nxt-1], d->enerflag);
			if (status != 0) {
				fprintf(stderr,"extract row: couldn't create row; failing\n");
				return -2;
			}
		} 
		for (n=0; n<nxt; n++) {
			r->data[DENSVAR][n] = d->data[DENSVAR][n][rowcol];
			r->data[ENERVAR][n] = d->data[ENERVAR][n][rowcol];
			r->data[PRESVAR][n] = d->data[PRESVAR][n][rowcol];
			r->data[TEMPVAR][n] = d->data[TEMPVAR][n][rowcol];
			r->data[MOMXVAR][n] = d->data[MOMXVAR][n][rowcol];
			r->data[MOMYVAR][n] = d->data[MOMYVAR][n][rowcol];
		}
		r->enerflag = d->enerflag;
	}

	if (direction == YDIR) {
		if (rowcol < 0 || rowcol >= nxt) {
			fprintf(stderr,"extract_row: invalid column %d; failing\n", rowcol);
			return -1;
		}

		if ((r->alloced != 0) && (r->N != nyt)) free_row(r);
		if (r->alloced == 0) {
			status = init_row(r, nyt, d->y[0], d->y[nyt-1], d->enerflag);
			if (status != 0) {
				fprintf(stderr,"extract row: couldn't create col; failing\n");
				return -2;
			}
		}
		
		for (n=0; n<nyt; n++) {
			r->data[DENSVAR][n] = d->data[DENSVAR][rowcol][n];
			r->data[ENERVAR][n] = d->data[ENERVAR][rowcol][n];
			r->data[PRESVAR][n] = d->data[PRESVAR][rowcol][n];
			r->data[TEMPVAR][n] = d->data[TEMPVAR][rowcol][n];
			r->data[MOMYVAR][n] = d->data[MOMXVAR][rowcol][n];
			r->data[MOMXVAR][n] = d->data[MOMYVAR][rowcol][n];
		}
		r->enerflag = d->enerflag;
	}

	return 0;
}

int domain_insert_row(domain_t *d, row_t *r, int rowcol, int direction) {

	int n;
	int nxt, nyt;

	nxt = d->Nx + 2*d->Nguard;
	nyt = d->Ny + 2*d->Nguard;

	switch (d->enerflag) {
		case FLAG_TOTALE: if (r->enerflag != d->enerflag) ener_internal_to_tot(r);
				  break;
		case FLAG_SPECIFICINTERNALE: if (r->enerflag != d->enerflag) ener_tot_to_internal(r);
				  break;
		default: fprintf(stderr,"domain_insert_row: d->enerflag messed up!  Continuing, but probably not good\n");
				  break;
	}

	if (direction != XDIR && direction != YDIR) {
		fprintf(stderr,"insert_row: invalid direction %d; failing\n", direction);
		return -1;
	}

	if (direction == XDIR) {
		if (rowcol < 0 || rowcol >= nyt) {
			fprintf(stderr,"insert_row: invalid row %d; failing\n", rowcol);
			return -1;
		}

		for (n=0; n<nxt; n++) {
			d->data[DENSVAR][n][rowcol] = r->data[DENSVAR][n];
			d->data[ENERVAR][n][rowcol] = r->data[ENERVAR][n];
			d->data[PRESVAR][n][rowcol] = r->data[PRESVAR][n];
			d->data[TEMPVAR][n][rowcol] = r->data[TEMPVAR][n];
			d->data[MOMXVAR][n][rowcol] = r->data[MOMXVAR][n];
			d->data[MOMYVAR][n][rowcol] = r->data[MOMYVAR][n];
		}
	}

	if (direction == YDIR) {
		if (rowcol < 0 || rowcol >= nxt) {
			fprintf(stderr,"extract_row: invalid column %d; failing\n", rowcol);
			return -1;
		}

		for (n=0; n<nyt; n++) {
			d->data[DENSVAR][rowcol][n] = r->data[DENSVAR][n];
			d->data[ENERVAR][rowcol][n] = r->data[ENERVAR][n];
			d->data[PRESVAR][rowcol][n] = r->data[PRESVAR][n];
			d->data[TEMPVAR][rowcol][n] = r->data[TEMPVAR][n];
			d->data[MOMYVAR][rowcol][n] = r->data[MOMXVAR][n];
			d->data[MOMXVAR][rowcol][n] = r->data[MOMYVAR][n];
		}
	}

	return 0;
}


