#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include "domain.h"
#include "get_options.h"
#include "util.h"
#include "kurganovlevy.h"
#include "domain_output.h"
#include "boundary_conds.h"
#include "init_conds.h"


int main(int argc, char **argv) {
    int npts  = 100;
    int nsteps = 500;
    double dt  = 5e-4;

    char ppmfilename[100];
    char h5filename[100];
    char ncdffilename[100];
    char binfilename[100];
    int xbc = BC_PERIODIC;
    int ybc = BC_PERIODIC;
    int outevery = 5;
    int output = 1;
    double psize;
    double pvel;
    double pdens = 50.;
    double pener;
    double time;
    double xmin, xmax; /* size of domain */
    double ymin, ymax; /* size of domain */
    double dx, dy;     /* resolution in domain */
    int step;
    int simulation = SIM_PROJECTILE;
    int outputvar;
    int bcs[4];
    int outnum = 0;
    int res;
    int nxproc=1, nyproc=1;
    

    domain_t d;
    struct timeval tt;

    res = get_options(argc,argv,&npts,&nsteps,&xbc,&ybc,&simulation,&outevery,&dt,&pener,&pvel,&psize,&output,&nxproc,&nyproc);
    if (res != 0) return 0;

    printf("Starting run with [%dx%d] grid, \n", npts, npts);
    printf("             for %d steps, each %g long \n", nsteps, dt);
    if (output) 
        printf("             outputting every %d steps\n", outevery);
    switch(simulation) {
        case SIM_PROJECTILE:
        printf("             with projectile of size %f, velocity %f, and density %f,\n", psize, pvel, pdens);
        printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
        break;
        case SIM_SEDOV:
        printf("             with explosion energy %g in size %g,\n", pener, psize);
        printf("             boundary conditions %d on x and %d on y\n", abs(xbc), abs(ybc));
        break;
    }

    /* setup ICs */
    switch(simulation) {
        case SIM_SEDOV: 
            ymin = xmin = -1.5;
            ymax = xmax = +1.5;
            dx = (xmax-xmin)/npts;
            dy = (ymax-ymin)/npts;
            init_domain(&d, npts, npts, KL_NGUARD, xmin, ymin, xmax, ymax);
            sedov_initvalues(&d, psize, pener);
            outputvar = PRESVAR;
            break;

        case SIM_PROJECTILE: 
            ymin = xmin = 0.;
            ymax = xmax = 1.;
            dx = (xmax-xmin)/npts;
            dy = (ymax-ymin)/npts;
            init_domain(&d, npts, npts, KL_NGUARD, xmin, ymin, xmax, ymax);
            projectile_initvalues(&d, psize, pdens, pvel);
            outputvar = DENSVAR;
            break;
    }

    /* apply boundary conditions and make thermodynamically consistant */
    bcs[0] = xbc; bcs[1] = xbc;
    bcs[2] = ybc; bcs[3] = ybc;
    apply_all_bcs(&d,bcs);
    domain_backward_dp_eos(&d);
    domain_ener_internal_to_tot(&d);  


    /* main loop */

    tick(&tt);
    if (output) domain_plot(&d);
    printf("Step\tdt\ttime\n");
    for (time=0.,step=0; step < nsteps; step++, time+=2.*dt) {

        printf("%d\t%g\t%g\n", step, dt, time);

        if (output && ((step % outevery) == 0) ) {
            sprintf(ppmfilename,"dens_test_%d.ppm", outnum); 
            sprintf(binfilename,"dens_test_%d.bin", outnum); 
            sprintf(h5filename,"dens_test_%d.h5", outnum); 
            sprintf(ncdffilename,"dens_test_%d.nc", outnum); 
            domain_output_ppm(&d, outputvar, ppmfilename);
            domain_output_bin(&d, binfilename);
            domain_output_hdf5(&d, h5filename);
            domain_output_netcdf(&d, ncdffilename);
            domain_plot(&d);
            outnum++;
        }
        kl_timestep_xy(&d, bcs, dt);
        apply_all_bcs(&d,bcs); 

        kl_timestep_yx(&d, bcs, dt);
        apply_all_bcs(&d,bcs); 
    }
    tock(&tt);

    /* final output */
    if (output) {
        sprintf(ppmfilename,"dens_test_%d.ppm", outnum); 
        domain_output_ppm(&d, outputvar,ppmfilename);
        domain_closeplot();
    }
    if (simulation == 0) domain_output_dens_r(&d,"sedov-test.dat");

    free_domain(&d);    
    return 0;
}


