 /*-------------------------------------------------------------------------
   lj.c: Lennard-Jones N-body simulation in C using MPI 
  -------------------------------------------------------------------------

   Input (from standard input or input file): 
       number of atoms  
       density of the system
       initial temperature (standard deviation of the velocities)
       runtime (from last check point)
       time step 
       random number generator seed
       equilibration time 
       number of time step after which to write xsf file (0=disable)
       checkpoint interval
       checkpoint filename

     Output (from standard output): separate lines containing:
       time, energy H, pot. en. U, kin. en. K, temperature T, fluctuations
     All reported energies values are divided by the number of atoms
     Fluctuations are the root mean square of E-<E>, with <E> the mean E

     Checkpoints are saved in the specified checkpoint filename, 
     and are read automatically if present.

     To compile:
       make

     To run:
       mpirun -np 8 lj input.ini > output.dat

     Ramses van Zon - SciNet - UoT - 2009/2010
  -------------------------------------------------------------------------*/

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "mpiglobal.h"
#include "parameters.h"
#include "system.h"

int main(int argc, char **argv)
{
    double start_time;
    double stop_time;
    FILE*  f;
    struct Parameters p;

    MPI_Init(&argc, &argv);
    MPI_global_init();     

    start_time = MPI_Wtime();

    if (argc == 1) {
        read_parameters(&p, MPI_COMM_WORLD, stdin);
    } else {
        if (rank==root) {
            f = fopen(argv[1], "r");
            if (!f) { 
                fprintf(stderr, "ERROR: Input file cannot be found\n");
                MPI_Abort(MPI_COMM_WORLD, 5);
            } 
        }
        read_parameters(&p, MPI_COMM_WORLD, f);
        if (rank==root)
            fclose(f);
    }

    printf("# proc%d: Creating system\n", rank);
    system_create(&p);

    printf("# proc%d: Running\n",rank);
    system_run();

    MPI_Barrier(MPI_COMM_WORLD);
    if (rank == root) {
        stop_time = MPI_Wtime();
        printf("#Wall clock time: %lf seconds\n", stop_time - start_time);
    }

    system_destroy();

    MPI_Finalize();

    return 0;
}
