#include <mpi.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "mpiglobal.h"
#include "system.h"
#include "latgen.h"
#include "communication.h"
#include "parameters.h"
#include "checkpoint.h"

#define NEW_ARRAY(size,type) (type*)malloc(sizeof(type)*(size))

int      n_tot;
long     seed;          
double   side;  
double   dt;            
double   runtime;       
double   temp;         
double   burn_in;        
double   cp_interval;    
char     cp_filename[MAXCPFILENAMELEN]; 
double   l[3]; 
double   origin[3];     
double   end[3];    
double   lc[3];   
int      n_max;          
int      n_cells;        
int      cell_dim;       
int      ncx, ncy, ncz; 
int      coord[3];      
int      proc_dim;       
int      np[3];         
int      neg_proc[3];    
int      pos_proc[3];    
int      n_max_neighbors; 
int      burn_in_steps;   
MPI_Comm comm;
int      n_steps;      
int      checkpoint_steps; 
int      per;              
int      N;
int      count;
int      n_points;
int      frame;         
double   e_kin;
double   e_pot;
double   e_tot;             
double   accum_e_tot;          
double   accum_e_tot2;         
struct Atom*     atom;   
struct IdVector* force;  
struct Atom*     buffer_atom; 
struct IdVector* outflow;      
struct IdVector* neighbor;     
struct IdVector* neighbor_force; 
int*             cell_list;      

const double rc  = 3.0;     
const double rs  = 2.5;     
const double rc2 = 3.0*3.0; 
const double rs2  = 2.5*2.5; 
const int offset[13][3] = {{ 1,0,0}, { 1,1,0},  {0, 1,0}, {-1,1,0},
                           { 0,0,1}, { 1, 0,1}, {1, 1,1}, { 0,1,1},
                           {-1,1,1}, {-1,0,1}, {-1,-1,1}, {0,-1,1},
                           { 1,-1,1}}; 


int file_exists(const char *s)
{
    FILE *f = fopen(s,"r");

    if (f == 0) 
        return 0;

    fclose(f);

    return 1; 
}
  

#define RM_SUCCESS 0
#define MV_SUCCESS 0
void backup(const char *name)
{
    if (file_exists(name)) {

        char new_name[strlen(name)+2];

        strcpy(new_name, name);
        strcat(new_name, "~");

        if (file_exists(new_name)) 
            if (remove(new_name) != RM_SUCCESS)
                printf("Warning: removal of previous backup %s failed\n", 
                       new_name);

        if (rename(name, new_name) != MV_SUCCESS)
            printf("Warning: backup of file %s failed\n", name);
    }
}

 
int min(int a, int b) 
{
    if (a > b) return b; else return a;
}

int max(int a, int b) 
{
    if (a > b) return a; else return b;
}

 
void divide_space()
{    
    coord[X] = coord[Y] = coord[Z] = 0;
    if (proc_dim)
        MPI_Cart_coords(comm, rank, proc_dim, coord);
     
    int d;
    for (d = 0; d < 3; d++) {
        l[d] = side / np[d];                     
        origin[d] = l[d] * coord[d] - 0.5*side;  
        end[d] = origin[d] + l[d];            
    }
     
    ncx = (int)(l[X]/rc);
    ncy = (int)(l[Y]/rc);
    ncz = (int)(l[Z]/rc);
     
    int no_cells = 0;
    if ( (np[X]==1 && ncx<3) || (np[X]>1 && ncx<2 ) ||
         (np[Y]==1 && ncy<3) || (np[Y]>1 && ncy<2 ) ||
         (np[Z]==1 && ncz<3) || (np[Z]>1 && ncz<2 ) )
        no_cells = 1;

    if (no_cells) {
        printf("# proc%d: No cells\n", rank);
        cell_dim = 0;  
    } else {
        cell_dim = 0; 
        lc[X] = l[X] / ncx;
        if (ncx > 1) 
            cell_dim++;
        lc[Y] = l[Y] / ncy;
        if (ncy > 1) 
            cell_dim++;
        lc[Z] = l[Z] / ncz;
        if (ncz > 1) 
            cell_dim++;
        printf("# proc%d: Cell geometry: (%d %d %d)\n"
               "# proc%d: Cell dimensions: (%lf %lf %lf)\n",
               rank, ncx, ncy, ncz,
               rank, lc[X], lc[Y], lc[Z]);
    }
}

 
void divide_processors(int npmaxx, int npmaxy, int npmaxz)
{
    float FnpUse = 0;
    float z = 1.0f/(npmaxx*npmaxx+npmaxy*npmaxy+npmaxz*npmaxz);
    int ix, iy, iz, d;

    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    for (iz=1; iz<=npmaxz; iz++) 
        for (iy=1; iy<=npmaxy; iy++) 
            for (ix=1; ix<=npmaxx; ix++) {
                float tryFnpUse = ix*iy*iz
                    -z*((ix-iy)*(ix-iy)+(iz-iy)*(iz-iy)+(ix-iz)*(ix-iz));
                if (tryFnpUse <= nprocs && tryFnpUse > FnpUse) {
                    FnpUse = tryFnpUse;
                    np[X] = ix;
                    np[Y] = iy;
                    np[Z] = iz;
                }
            }

    proc_dim = 0;
    for (d=0; d<3; d++)
        if (np[d] > 1) 
            proc_dim++;

    if (np[X] * np[Y] * np[Z] != nprocs) {
        if (rank == root) {
            fprintf(stderr,"There is only use for %d  processors ###\n\n",
                    np[X] * np[Y] * np[Z]);
            fprintf(stderr, "ERROR: Reduce the number of processors!");
        }
        MPI_Abort(MPI_COMM_WORLD,6);	
    }
    
    int periods[3] = {1,1,1};
    if (proc_dim) 
        MPI_Cart_create(MPI_COMM_WORLD, proc_dim, np, periods, 0, &comm);
    else
        MPI_Comm_dup(MPI_COMM_WORLD, &comm);
     
    MPI_Comm_rank(comm, &rank);

    if (rank==root) 
        printf("# (np[X] np[Y] np[Z]) = (%d %d %d)\n", np[X], np[Y], np[Z]);

     
    for (d=0; d<proc_dim; d++) 
        MPI_Cart_shift(comm, d, 1, &(neg_proc[d]), &(pos_proc[d]));
}

 
int lin_cell(int cx, int cy, int cz) 
{
    if (cx < 0) 
        cx += ncx;
    else if (cx >= ncx) 
        cx -= ncx;
    if (cy< 0) 
        cy+= ncy;
    else if (cy>= ncy) 
        cy -= ncy;
    if (cz < 0) 
        cz += ncz;
    else if (cz >= ncz) 
        cz -= ncz;
    return (cz*ncy+cy)*ncx+cx;
}

 
double gaussian()
{
    static int have = 0;
    static double x2;
    double fac, y1, y2, x1;

    if (have == 1) {   
        have = 0;
        return x2;
    } else {
         
        y1 = drand48();
        y2 = drand48();
        fac = sqrt(-2*log(y1));
        have = 1;
        x1 = fac*sin(2*M_PI*y2);  
        x2 = fac*cos(2*M_PI*y2);  
        return x1;                
    }
}

void redistribute()
{
    if (nprocs == 1) {

        memcpy(atom, buffer_atom, N*sizeof(struct Atom));

    } else {        

        int  n_read = N;
        int  n2move[nprocs];  
        int* elmoff[nprocs];  
                
        int i, p;
        for (p = 0; p < nprocs; p++) {
            n2move[p] = 0;
            elmoff[p] = NEW_ARRAY(n_max,int);
        }
         
        N = 0;
         
        int tot_send = 0;
        for (i = 0; i < n_read; i++) {
            int coords[3] = { (buffer_atom[i].q[X] + 0.5*side)/l[0],
                              (buffer_atom[i].q[Y] + 0.5*side)/l[1],
                              (buffer_atom[i].q[Z] + 0.5*side)/l[2] };
            int irank;
            MPI_Cart_rank(comm, coords, &irank);
            if (irank == rank)
                atom[N++] = buffer_atom[i];            
            else {
                elmoff[irank][n2move[irank]++] = i;   
                tot_send++;
            }
        }
        
        if (tot_send != 0)
	        printf("# %d atoms out of %d need to be moved off of proc %d\n", 
                   tot_send, n_read, rank);

        int* elmsize = NEW_ARRAY(n_max,int);  
        for (i = 0; i < n_max; i++) 
            elmsize[i] = 1;
                 
        int tot_n_receive = nprocs-1;  
        int n_sent = 0;                     
        MPI_Request request[tot_n_receive];          
         
        for (p = 0; p < nprocs; p++) 
            if (p != rank) {
                int send_first = (p+rank)&1;
                if (p<rank)
                    send_first = !send_first;
                if (send_first) {
                    MPI_Datatype send_type;
                    MPI_Type_indexed(n2move[p], elmsize, elmoff[p], MPI_ATOM, 
                                     &send_type);
                    MPI_Type_commit(&send_type);
                    MPI_Isend(buffer_atom, 1, send_type, p, 15, comm, 
                              request+n_sent);
                    MPI_Type_free(&send_type);
                    n_sent++; 
                }
            }
                 
        int n_receive1 = tot_n_receive - n_sent;   
        int m;
        for (m = 0; m < n_receive1; m++) {
            MPI_Status status;
            MPI_Recv(atom+N, n_max-N, MPI_ATOM, MPI_ANY_SOURCE, 15, comm, &status);
            int dn=0;
            MPI_Get_count(&status, MPI_ATOM, &dn);
            N += dn;
        }
                 
        for (p = 0; p < nprocs; p++) {
            if (p!=rank) {
                int send_first = (p+rank)&1;
                if (p<rank)
                    send_first = !send_first;
                int send_last = !send_first;
                if (send_last) {
                    MPI_Datatype send_type;
                    MPI_Type_indexed(n2move[p], elmsize, elmoff[p], MPI_ATOM, 
                                     &send_type);
                    MPI_Type_commit(&send_type);
                    MPI_Isend(buffer_atom, 1, send_type, p, 15, comm, 
                              request+n_sent);
                    MPI_Type_free(&send_type);
                    n_sent++;
                }
            }
        }
                 
        int n_receive2 = tot_n_receive - n_receive1;  
        for (m = 0; m < n_receive2; m++) {
            MPI_Status status;
            MPI_Recv(atom+N, n_max-N, MPI_ATOM, MPI_ANY_SOURCE, 15, comm, 
                     &status);
            int dn = 0;
            MPI_Get_count(&status, MPI_ATOM, &dn);
            N += dn;
        }

        MPI_Barrier(comm);  
        for (p = 0; p < nprocs; p++) 
            free(elmoff[p]);
        free(elmsize);
    }
}


double phi(double a[3], double b[3], double fa[3], double fb[3]) 
{
    double eij = 0.0;    
    double dr[3]; 
    int d;

    dr[X] = a[X] - b[X];  
    dr[Y] = a[Y] - b[Y];  
    dr[Z] = a[Z] - b[Z];  

    for (d=proc_dim; d<3; d++) 
        if (dr[d] < -0.5*side) dr[d] += side;     
        else if (dr[d] >= 0.5*side) dr[d] -= side;  

    double r2 = dr[X]*dr[X] + dr[Y]*dr[Y] + dr[Z]*dr[Z];  
    if (r2 < rc2) {          
        double r2i = 1/r2;
        double r6i = r2i*r2i*r2i;
        double fij = 48*r2i*r6i*(r6i-0.5);  
        eij = 4*r6i*(r6i-1);  
        if (r2 > rs2) {             
            double r = sqrt(r2);
            double x = (2*r-rs-rc)/(rs-rc);
            double x2 = x*x;
            double alpha = 0.5-0.25*x*(x2-3);
            double dalpha = 1.5*(x2-1)/(r*(rs-rc));
            fij = alpha*fij+dalpha*eij;
            eij = alpha*eij;
        }
         
        fa[X] += fij*dr[X];
        fa[Y] += fij*dr[Y];
        fa[Z] += fij*dr[Z];
        fb[X] -= fij*dr[X];
        fb[Y] -= fij*dr[Y];
        fb[Z] -= fij*dr[Z];
    }
    return eij;
}


void compute_forces()
{ 
    int d, i, j, k, n_inflow1, n_inflow2, n_neighbors=0;
     
    e_pot = 0; 
    for (k=0; k<N; k++) 
        force[k].v[X] = force[k].v[Y] = force[k].v[Z] = 0;
     
    if (cell_dim) {
         
        for (k=N; k < N+ncx*ncy*ncz; k++) 
            cell_list[k] = -1;  
        for (k=0; k<N; k++) {             
            int cix = (int)((atom[k].q[X]-origin[X])/lc[X]);
            int ciy = (int)((atom[k].q[Y]-origin[Y])/lc[Y]);
            int ciz = (int)((atom[k].q[Z]-origin[Z])/lc[Z]);
            int c = lin_cell(cix,ciy,ciz) + N;    
            cell_list[k] = cell_list[c];  
            cell_list[c] = k;
        }

        int icx, icy, icz;
        for (icx=0; icx<ncx; (icx)++)  {
            for (icy=0; icy<ncy; (icy)++) { 
                for (icz=0; icz<ncz; (icz)++) {
                    int c = lin_cell(icx,icy,icz) + N;
                     
                    for (i=cell_list[c]; i>=0; i=cell_list[i]) 
                        for (j=cell_list[c]; j>=0; j=cell_list[j]) 
                            if (i < j) 
                                e_pot += phi(atom[i].q,atom[j].q,force[i].v,force[j].v); 
                     
                    for (k=0; k < n_send_tag[cell_dim]; ++k) {
                         
                        int jcx = icx + offset[k][2];
                        int jcy = icy + offset[k][1];
                        int jcz = icz + offset[k][0];
                        if ( ( np[X]==1 || (jcx>=0 && jcx<ncx) )
                             &&
                             ( np[Y]==1 || (jcy>=0 && jcy<ncy) )
                             &&
                             ( np[Z]==1 || (jcz>=0 && jcz<ncz) ) )
                        {
                            int cj = lin_cell(jcx,jcy,jcz) + N;
                            for (i=cell_list[c]; i>=0; i=cell_list[i]) 
                                for (j=cell_list[cj]; j>=0; j=cell_list[j]) 
                                    e_pot += phi(atom[i].q,atom[j].q,force[i].v,force[j].v);
                        }
                    }
                }
            }
        }

    } else {
         
        for (i=0; i<N; ++i) 
            for (j=0; j<i; ++j) 
                e_pot += phi(atom[i].q,atom[j].q,force[i].v,force[j].v);
    }

    for (i=0; i<N; ++i) {
        atom[i].tag = nowhere; 
        for (d=0; d<proc_dim; ++d) {
            if (atom[i].q[d] < origin[d]+rc) atom[i].tag ^= neg_tag[d];
            if (atom[i].q[d] >= end[d] - rc) atom[i].tag ^= pos_tag[d];
        }
    }

     
    n_neighbors = 0;
    for (d=0; d<proc_dim; ++d) {
        int n_outflow = 0;
        Atom_pack(pos_tag[d], N, atom, &n_outflow, outflow);
        if (d > 0)
            IdVector_pack(pos_tag[d], n_neighbors, neighbor, &n_outflow, outflow);
        if (coord[d] == np[d]-1) 
            for (i=0; i<n_outflow; ++i) 
                outflow[i].v[d] -= side;
        struct IdVector* inflow1 = neighbor + n_neighbors;
        IdVector_exchange(&n_inflow1, inflow1, neg_proc[d],
                          n_outflow, outflow, pos_proc[d],
                          16, 
                          n_max_neighbors-n_neighbors,
                          comm);
        if (d==proc_dim-1) {
            n_neighbors += n_inflow1;
        } else {
            n_outflow = 0;
            Atom_pack(neg_tag[d], N, atom, &n_outflow, outflow);
            if (d > 0)
                IdVector_pack(neg_tag[d],n_neighbors,neighbor,&n_outflow,outflow);
            if (coord[d] == 0) 
                for (i=0; i<n_outflow; ++i) 
                    outflow[i].v[d] += side;
            struct IdVector* inflow2 = inflow1 + n_inflow1;
            IdVector_exchange(&n_inflow2, inflow2, pos_proc[d],
                              n_outflow, outflow, neg_proc[d],
                              17,
                              n_max_neighbors-n_neighbors-n_inflow1,
                              comm);
            n_neighbors += n_inflow1 + n_inflow2;
        }
    }

    for (j=0; j<n_neighbors; ++j) {
        i = n_send_tag[proc_dim];
        while (i--) 
            if (neighbor[j].route == send_tag[i]) 
                break;
        if (i < 0) 
            neighbor[j].tag = to_be_removed;
    }
    IdVector_prune(&n_neighbors, neighbor);
    for (j=0; j<n_neighbors; ++j) {
        neighbor_force[j].v[X]=neighbor_force[j].v[Y]=neighbor_force[j].v[Z]=0;
        neighbor_force[j].localid = neighbor[j].localid;
        neighbor_force[j].tag = reverse_tag(neighbor[j].route);
    }

    if (cell_dim == 0) {

        for (j=0; j<n_neighbors; ++j) 
            for (i=0; i<N; ++i)
                e_pot += phi(atom[i].q,neighbor[j].v,force[i].v,neighbor_force[j].v);

    } else {
         
        for (j=0; j<n_neighbors; ++j) {
             
            int jcx = (int)((neighbor[j].v[X]-origin[X])/lc[X]);
            int jcy = (int)((neighbor[j].v[Y]-origin[Y])/lc[Y]);
            int jcz = (int)((neighbor[j].v[Z]-origin[Z])/lc[Z]);
             
            int minicx = np[X]==1?(jcx-1):max(jcx-1,0);
            int minicy = np[Y]==1?(jcy-1):max(jcy-1,0);
            int minicz = np[Z]==1?(jcz-1):max(jcz-1,0);
            int maxicz = np[Z]==1?(jcz+1):min(jcz+1,ncz-1);
            int maxicx = np[X]==1?(jcx+1):min(jcx+1,ncx-1);
            int maxicy = np[Y]==1?(jcy+1):min(jcy+1,ncy-1);
             
            if (neighbor_force[j].tag & negX)      minicx = maxicx = 0;
            else if (neighbor_force[j].tag & posX) minicx = maxicx = ncx-1;
            if (neighbor_force[j].tag & negY)      minicy = maxicy = 0;
            else if (neighbor_force[j].tag & posY) minicy = maxicy = ncy-1;
            if (neighbor_force[j].tag & negZ)      minicz = maxicz = 0;
            else if (neighbor_force[j].tag & posZ) minicz = maxicz = ncz-1;
             
            int icx, icy, icz;
            for (icx=minicx; icx<=maxicx; ++icx) 
                for (icy=minicy; icy<=maxicy; ++icy) 
                    for (icz=minicz; icz<=maxicz; ++icz) 
                        for (i=cell_list[lin_cell(icx,icy,icz)+N]; 
                             i>=0; 
                             i=cell_list[i])
                            e_pot += phi(atom[i].q, neighbor[j].v, 
				     force[i].v, neighbor_force[j].v);

        }
    }
    
    int n_neighbor_forces = 0;
    for (d=0; d<proc_dim; ++d) {
        int n_outflow = 0;
        IdVector_move(pos_tag[d], &n_neighbors, neighbor_force, &n_outflow, outflow);
        IdVector_move(pos_tag[d], &n_neighbor_forces, neighbor, &n_outflow, outflow);
        struct IdVector* inflow = neighbor + n_neighbor_forces;
        int n_inflow;
        IdVector_exchange(&n_inflow, inflow, neg_proc[d],
                          n_outflow, outflow, pos_proc[d],
                          16,
                          n_max_neighbors-n_neighbor_forces,
                          comm);
        n_outflow = 0;
        IdVector_move   (neg_tag[d], &n_neighbors, neighbor_force, &n_outflow, outflow);
        IdVector_tagMove(neg_tag[d], &n_neighbor_forces, neighbor, &n_outflow, outflow);
        n_neighbor_forces += n_inflow;
        IdVector_prune(&n_neighbor_forces,neighbor);
        inflow = neighbor + n_neighbor_forces;
        IdVector_exchange(&n_inflow, inflow, pos_proc[d],
                          n_outflow, outflow, neg_proc[d],
                          17,
                          n_max_neighbors-n_neighbor_forces,
                          comm);
        n_neighbor_forces += n_inflow;
    }
     
    for (i=0; i<n_neighbor_forces; ++i) {
        force[neighbor[i].localid].v[X] += neighbor[i].v[X];
        force[neighbor[i].localid].v[Y] += neighbor[i].v[Y];
        force[neighbor[i].localid].v[Z] += neighbor[i].v[Z];
    }
}

 
void initialize() 
{
    burn_in_steps = (int)(burn_in/dt+0.5);
    checkpoint_steps = (int)(cp_interval/dt+0.5);
    n_points = 0;
    accum_e_tot = 0; 
    accum_e_tot2 = 0;
    count = 0;
    frame = 0; 

    if (file_exists(cp_filename)) {
         
        if (rank==root)
            printf("# reading checkpoint file %s\n", cp_filename);

        cp_read(buffer_atom, &N, n_tot, MPI_ATOM, cp_filename, comm);

        redistribute();

    } else {
         
        struct Generator gen;
        double scale = sqrt(temp); 
        double a = side/(int)(cbrt((double)(n_tot))+0.99999999999); 
        int n_possible, n, d;
        init(&gen, origin, end, side, a);
        n_possible = count_lattice_positions(&gen);
        N = 0;
        for (n=0; n<n_possible; n++) {
            int i = make_lattice_position(&gen);
            if (i < n_tot) {
                atom[N].id = i;
                atom[N].q[X] = gen.r[X];
                atom[N].q[Y] = gen.r[Y];
                atom[N].q[Z] = gen.r[Z];
                (N)++;
            } else
                break;
        }
         
        srand48(seed); 
        
        for (n=0; n<N; n++) 
            for (d=0; d<3; d++)
                atom[n].p[d] = scale*gaussian();
    }

     
    e_kin = 0;
    int n;
    for (n = 0; n < N; n++) 
        e_kin += atom[n].p[X]*atom[n].p[X] 
           + atom[n].p[Y]*atom[n].p[Y] 
           + atom[n].p[Z]*atom[n].p[Z];   
    temp = e_kin/(3*N);
    e_kin /= 2;

    compute_forces();

    n_steps = (int)(runtime/dt+0.5); 
     
    double e_kin_tot, e_pot_tot;
    MPI_Reduce(&e_kin,&e_kin_tot,1,MPI_DOUBLE,MPI_SUM,root,comm);
    MPI_Reduce(&e_pot,&e_pot_tot,1,MPI_DOUBLE,MPI_SUM,root,comm);

    if (rank == root)
        printf("# done initializing\n");

    if (rank == root) {
        e_kin_tot /= n_tot;
        e_pot_tot /= n_tot;
        printf("# t        E        U        K        T      <[H-<H>]^2>\n");
        printf("%8.5f %8.5f %8.5f %8.5f %8.5f ", 
               count*dt, e_pot_tot+e_kin_tot, e_pot_tot, e_kin_tot, e_kin_tot/1.5);
        if (count < burn_in_steps) {
            printf("0.000\n");
        } else {
            double avgH = accum_e_tot/n_points;
            double avgH2 = accum_e_tot2/n_points;
            double fluctH = sqrt(fabs(avgH2-avgH*avgH));
            printf("%e\n", fluctH);
        }
    }
}

 
void integrate_step()
{
    int i, d;
    
    for (i=0; i<N; ++i) {
        atom[i].p[X] += 0.5*dt*force[i].v[X];
        atom[i].p[Y] += 0.5*dt*force[i].v[Y];
        atom[i].p[Z] += 0.5*dt*force[i].v[Z];
    }
   
    for (i=0; i<N; ++i) {
        atom[i].q[X] += dt*atom[i].p[X];
        atom[i].q[Y] += dt*atom[i].p[Y];
        atom[i].q[Z] += dt*atom[i].p[Z];
    }
       
    for (i=0; i<N; ++i) {
        atom[i].tag = nowhere;
        for (d=0; d<3; ++d) {
            if (atom[i].q[d] < origin[d])    atom[i].tag ^= neg_tag[d];
            else if (atom[i].q[d] >= end[d]) atom[i].tag ^= pos_tag[d];
        }
    }
     
    for (d=0; d<proc_dim; ++d) {
        int n_inflow;
        int n_outflow = 0;
         
        Atom_move(pos_tag[d], &N, atom, &n_outflow, buffer_atom);
         
        if (coord[d] == np[d]-1) 
            for (i=0; i<n_outflow; ++i) 
                buffer_atom[i].q[d] -= side;
         
        Atom_exchange(&n_inflow, atom+N,      neg_proc[d],
                      n_outflow, buffer_atom, pos_proc[d],
                      pos_tag[d], n_max - N, comm);
         
        N += n_inflow;
         
        n_outflow = 0;
        Atom_move(neg_tag[d], &N, atom, &n_outflow, buffer_atom);
        if (coord[d] == 0)
            for (i=0; i<n_outflow; ++i) 
                buffer_atom[i].q[d] += side;
        Atom_exchange(&n_inflow, atom+N,      pos_proc[d],
                      n_outflow, buffer_atom, neg_proc[d],
                      neg_tag[d], 
                      n_max - N, comm);
        N += n_inflow;
    }
     
    for (d=proc_dim; d<3; ++d) {
        for (i=0; i<N; ++i) 
            if (atom[i].tag & pos_tag[d]) 
                atom[i].q[d] -= side;
            else if (atom[i].tag & neg_tag[d]) 
                atom[i].q[d] += side;
    }
     
    compute_forces();
     
    e_kin = 0;
    for (i=0; i<N; ++i) {
        atom[i].p[X] += 0.5*dt*force[i].v[X];
        atom[i].p[Y] += 0.5*dt*force[i].v[Y];
        atom[i].p[Z] += 0.5*dt*force[i].v[Z];
        e_kin += atom[i].p[X]*atom[i].p[X] 
               + atom[i].p[Y]*atom[i].p[Y] 
               + atom[i].p[Z]*atom[i].p[Z];
    }
     
    temp = e_kin/(3*N);
    e_kin /= 2;
    e_tot = e_pot + e_kin;
}


void memfill(char*s, size_t sz, char c1, char c2)
{
   size_t ss = strlen(s), i;
   for (i=ss;i<sz;i++)
      s[i]=c1;
   s[sz-1]=c2;
}
 
void output_positions() 
{
    const int XSFHEADERSIZE = 64;
    const int XSFLINESIZE = 32;
    MPI_File mf;
    char     name[13];
    int      i;
    int      err;
     
    sprintf(name,"pos%03d.xsf",rank);
     
    if (frame++==0)    

        if (rank == root) {
             
            if (file_exists("pos.xsf"))
                MPI_File_delete("pos.xsf", MPI_INFO_NULL);
            MPI_File_open(MPI_COMM_SELF, "pos.xsf", 
                          MPI_MODE_CREATE|MPI_MODE_WRONLY,
                          MPI_INFO_NULL, &mf);
            char header[XSFHEADERSIZE];
            sprintf(header,
                    "PRIMVEC\n"
                    "%8.5lf 0.0 0.0\n"
                    "0.0 %8.5lf 0.0\n"
                    "0.0 0.0 %8.5lf",
                    side,side,side);
            memfill(header,XSFHEADERSIZE,' ','\n');
            err = MPI_File_write(mf,header,XSFHEADERSIZE,MPI_CHAR, 
                                 MPI_STATUS_IGNORE);
            MPI_File_close(&mf);    
        }

    MPI_File_open(comm, "pos.xsf", MPI_MODE_APPEND|MPI_MODE_WRONLY, 
                  MPI_INFO_NULL, &mf);  
     
    char linebuffer[XSFLINESIZE];
    if (rank == root) {
        sprintf(linebuffer, "ATOMS %8d", frame);
        memfill(linebuffer,XSFLINESIZE,' ','\n');
        MPI_File_write_shared(mf, linebuffer, XSFLINESIZE, MPI_CHAR, 
                              MPI_STATUS_IGNORE);
    } else
        MPI_File_write_shared(mf, linebuffer, 0, MPI_CHAR, 
                              MPI_STATUS_IGNORE);

    for (i=0; i<n_max; ++i) {
        int j = i<N?i:N-1;
        sprintf(linebuffer, 
                "%c %8.5lf %8.5lf %8.5lf", 
                (char)('A'+rank), atom[j].q[X], atom[j].q[Y], atom[j].q[Z]);
        memfill(linebuffer,32,' ','\n');
        MPI_File_write_ordered(mf, linebuffer, XSFLINESIZE, MPI_CHAR,
                               MPI_STATUS_IGNORE);
    }
     
    MPI_File_close(&mf);
}

 
void system_run()
{     
    initialize();

    for (; count<n_steps; ++count) {
             
        integrate_step();
         
        double e_kin_tot, e_pot_tot;
        MPI_Reduce(&e_kin,&e_kin_tot,1,MPI_DOUBLE,MPI_SUM,root,comm);
        MPI_Reduce(&e_pot,&e_pot_tot,1,MPI_DOUBLE,MPI_SUM,root,comm);
        if (rank==root) {
            
            e_kin_tot /= n_tot;
            e_pot_tot /= n_tot;
            double e_tot_tot = e_kin_tot + e_pot_tot;
            printf("%8.5f %8.5f %8.5f %8.5f %8.5f ", 
                   (count+1)*dt, e_tot_tot, e_pot_tot, e_kin_tot, e_kin_tot/1.5);
            if (count < burn_in_steps) {

                printf("0.000\n");

            } else {
                 
                accum_e_tot += e_tot_tot;
                accum_e_tot2 += e_tot_tot*e_tot_tot;
                ++n_points;
                double avgH = accum_e_tot/n_points;
                double avgH2 = accum_e_tot2/n_points;
                double fluctH = sqrt(fabs(avgH2-avgH*avgH));
                printf("%e\n", fluctH);
            }
        }
                 
        if (per != 0 && count%per == 0) {
            if (rank==root)
                printf("# writing to xsf file pos.xsf\n");
            output_positions();
        }
         
        if (count%checkpoint_steps==0 && count>0) { 
            if (rank==root) 
                backup(cp_filename);
            MPI_Barrier(comm);
            if (rank==root)
               printf("# writing checkpoint file %s\n", cp_filename);
            cp_write(atom, N, MPI_ATOM, cp_filename, comm);
        }

    }
     
    if (rank==root) 
        backup(cp_filename);
    MPI_Barrier(comm);
    if (rank==root)
       printf("# writing checkpoint file %s\n", cp_filename);
    cp_write(atom, N, MPI_ATOM, cp_filename, comm);
}

 
void system_create(struct Parameters *p)
{    
    n_tot       = p->n_tot;
    seed        = p->seed;
    side        = p->side;  
    dt          = p->dt;
    runtime     = p->runtime;
    temp        = p->temp;      
    burn_in     = p->burn_in; 
    per         = p->per;
    cp_interval = p->cp_interval;
    strcpy(cp_filename, p->cp_filename);
     
    IdVector_commit();
    Atom_commit();

    if (side < 2*rc) { 
        fprintf(stderr,"System sizes < 2x the cut-off are not supported.");
        MPI_Abort(MPI_COMM_WORLD,7);
    }
     
    int npmax = (int)(side/rc);
    divide_processors(npmax,npmax,npmax);
     
    divide_space();

    if (rank==root) 
        printf("# Geometry of regions: (%d %d %d)\n"
               "# Each region is %lf %lf %lf\n",
               np[X], np[Y], np[Z],
               l[X], l[Y], l[Z]);
     
    double rho = n_tot/(side*side*side);
     
    n_max = (int)(13+1.5*l[X]*l[Y]*l[Z]*rho);    
    n_max_neighbors = (int)(3+18*(l[X]*l[Y]+l[X]*l[Z]+l[Y]*l[Z])*rc*rho);
     
    atom           = NEW_ARRAY(n_max,struct Atom);
    buffer_atom    = NEW_ARRAY(n_max,struct Atom);
    force          = NEW_ARRAY(n_max,struct IdVector);
    neighbor       = NEW_ARRAY(n_max_neighbors,struct IdVector);
    neighbor_force = NEW_ARRAY(n_max_neighbors,struct IdVector);
    outflow        = NEW_ARRAY(n_max_neighbors,struct IdVector);
    cell_list      = NEW_ARRAY((n_max+ncx*ncy*ncz),int);
}

 
void system_destroy()
{
    free(atom);
    free(force);
    free(cell_list);
    free(neighbor);
    free(neighbor_force);
    free(outflow);
    free(buffer_atom);
}

